import dis
import math
import random
import pygame
from .renderer import Renderer3D
from . import shapes

class ShapePhysicsObject:
    def __init__(self, renderer: Renderer3D, shape: str, rotation: tuple[float], color: tuple[int], size: float, mass: float, grid_size: float):
        self.renderer = renderer
        valid_shapes = ["sphere", "plane"]
        if shape not in valid_shapes:
            raise ValueError(f"Invalid Input '{shape}'. expected one of {valid_shapes}")
        
        self.shape = shape
        self.mass = mass
        self.grid_size = grid_size
        self.grid_coords = None
        self.rotation = rotation

        self.color = color

        self.forces = [0, 0, 0]
        self.anchor_position = [0, 0, 0]
        self.velocity = [0, 0, 0]
        self.timeStep = 1
        self.size = size

        if shape == "sphere":
            self.grid_coords = shapes.generate_sphere(resolution=grid_size, radius=size)
        elif shape == "plane":
            self.mass = float("inf")
            self.grid_coords = shapes.generate_plane(size=size, rot_x=math.radians(rotation[0]), rot_y=math.radians(rotation[1]), rot_z=math.radians(rotation[2]))
            #self.anchor_position = self.rotate_xyz(self.anchor_position, math.radians(rotation[0]), math.radians(rotation[1]), math.radians(rotation[2]))


            # --- Compute plane normal from first 3 vertices ---
            p0 = self.grid_coords[0][0]
            p1 = self.grid_coords[0][1]
            p2 = self.grid_coords[1][0]

            # Edge vectors
            e1 = (p1[0]-p0[0], p1[1]-p0[1], p1[2]-p0[2])
            e2 = (p2[0]-p0[0], p2[1]-p0[1], p2[2]-p0[2])

            # Cross product
            nx = e1[1]*e2[2] - e1[2]*e2[1]
            ny = e1[2]*e2[0] - e1[0]*e2[2]
            nz = e1[0]*e2[1] - e1[1]*e2[0]

            length = math.sqrt(nx*nx + ny*ny + nz*nz)
            if length != 0:
                nx /= length
                ny /= length
                nz /= length

            self.plane_normal = (nx, ny, nz)
            self.plane_point = p0

    def rotate_xyz(self, pos, rx, ry, rz):
        x, y, z = pos

        # Rotate X
        cosx, sinx = math.cos(rx), math.sin(rx)
        y, z = y*cosx - z*sinx, y*sinx + z*cosx

        # Rotate Y
        cosy, siny = math.cos(ry), math.sin(ry)
        x, z = x*cosy + z*siny, -x*siny + z*cosy

        # Rotate Z
        cosz, sinz = math.cos(rz), math.sin(rz)
        x, y = x*cosz - y*sinz, x*sinz + y*cosz

        return [x, y, z]

    def add_forces(self, force: tuple[float]):
        self.forces[0] += force[0]
        self.forces[1] += force[1]
        self.forces[2] += force[2]

    def set_forces(self, force: tuple[float]):
        self.forces[0] = force[0]
        self.forces[1] = force[1]
        self.forces[2] = force[2]

    def update_velocity_from_forces(self):
        last_velocity = self.velocity[:]
        self.velocity[0] = last_velocity[0] + (self.forces[0]/self.mass) * self.timeStep
        self.velocity[1] = last_velocity[1] + (self.forces[1]/self.mass) * self.timeStep
        self.velocity[2] = last_velocity[2] + (self.forces[2]/self.mass) * self.timeStep
        
    def update_pos_from_velocity(self):
        last_anchor_pos = self.anchor_position[:]
        self.anchor_position[0] = last_anchor_pos[0] + self.velocity[0]
        self.anchor_position[1] = last_anchor_pos[1] + self.velocity[1]
        self.anchor_position[2] = last_anchor_pos[2] + self.velocity[2]

    def apply_anchor_position_to_grid_coords(self):
        new_coords = []
        for y in range(len(self.grid_coords)):
            yGrid = self.grid_coords[y]
            new_coords1 = []
            for x in range(len(yGrid)):

                new_coords1.append(
                    (
                        self.grid_coords[y][x][0] + self.anchor_position[0],
                        self.grid_coords[y][x][1] + self.anchor_position[1],
                        self.grid_coords[y][x][2] + self.anchor_position[2]
                    )
                )

            new_coords.append(new_coords1)
        
        return new_coords

    def add_shape_to_renderer(self):
        new_coords = self.apply_anchor_position_to_grid_coords()
        self.renderer.grid_coords_list.append((new_coords, True))

    def add_color_to_renderer(self):
        self.renderer.triangle_color_list_1.append(self.color)
        self.renderer.triangle_color_list_2.append((self.color[0] * 0.75, self.color[1] * 0.75, self.color[2] * 0.75))

    def detect_collision(self, shapes: list["ShapePhysicsObject"]):
        if self.shape == "plane":
            return
        #print(shapes)
        for shapeI in range(len(shapes)):
            #print(shapeI)
            shape = shapes[shapeI]

            if shape.shape == "sphere":
                this_center_point = (self.anchor_position[0], self.anchor_position[1] - self.size, self.anchor_position[2])
                other_center_point = (shape.anchor_position[0], shape.anchor_position[1] - shape.size, shape.anchor_position[2])
                
                dx = this_center_point[0]-other_center_point[0]
                dy = this_center_point[1]-other_center_point[1]
                dz = this_center_point[2]-other_center_point[2]

                dist = math.sqrt(math.pow(dx, 2) + math.pow(dy, 2) + math.pow(dz, 2))

                if dist <= self.size + shape.size:
                    if dist == 0:
                        continue
                    nx = dx / dist
                    ny = dy / dist
                    nz = dz / dist
                    self.handle_collision(shape, (nx, ny, nz))
            
            elif shape.shape == "plane":
                # Sphere center
                center = (
                    self.anchor_position[0],
                    self.anchor_position[1] - self.size,
                    self.anchor_position[2]
                )

                # Use plane's anchor_position and normal
                plane_world_point = (
                    shape.plane_point[0] + shape.anchor_position[0],
                    shape.plane_point[1] + shape.anchor_position[1],
                    shape.plane_point[2] + shape.anchor_position[2]
                )
                n = shape.plane_normal

                px = center[0] - plane_world_point[0]
                py = center[1] - plane_world_point[1]
                pz = center[2] - plane_world_point[2]
                dist = px*n[0] + py*n[1] + pz*n[2]

                # print("shape: ", shape.shape, "dist: ", dist)

                if abs(dist) <= self.size:
                    vn = self.velocity[0]*n[0] + \
                        self.velocity[1]*n[1] + \
                        self.velocity[2]*n[2]

                    if dist * vn < 0:
                        # Reflect
                        self.velocity[0] -= 2 * vn * n[0]
                        self.velocity[1] -= 2 * vn * n[1]
                        self.velocity[2] -= 2 * vn * n[2]

                        # Positional correction
                        penetration = self.size - abs(dist)
                        sign = 1 if dist > 0 else -1

                        self.anchor_position[0] += n[0] * penetration * sign
                        self.anchor_position[1] += n[1] * penetration * sign
                        self.anchor_position[2] += n[2] * penetration * sign
            
    def handle_collision(self, other_shape: "ShapePhysicsObject", n):
        if other_shape.shape == "sphere":
            rvx = self.velocity[0] - other_shape.velocity[0]
            rvy = self.velocity[1] - other_shape.velocity[1]
            rvz = self.velocity[2] - other_shape.velocity[2]

            # --- Velocity along normal (dot product) ---
            vel_along_normal = rvx*n[0] + rvy*n[1] + rvz*n[2]

            # If objects are separating, do nothing
            if vel_along_normal > 0:
                return

            m1 = self.mass
            m2 = other_shape.mass

            # --- Compute impulse scalar ---
            j = (2 * vel_along_normal) / (m1 + m2)

            # --- Store old velocities ---
            v1x, v1y, v1z = self.velocity
            v2x, v2y, v2z = other_shape.velocity

            # --- Apply impulse ---
            self.velocity[0] = v1x - j * m2 * n[0]
            self.velocity[1] = v1y - j * m2 * n[1]
            self.velocity[2] = v1z - j * m2 * n[2]

            other_shape.velocity[0] = v2x + j * m1 * n[0]
            other_shape.velocity[1] = v2y + j * m1 * n[1]
            other_shape.velocity[2] = v2z + j * m1 * n[2]
        elif other_shape.shape == "plane":
            n = other_shape.plane_normal

            # Dot product
            vn = self.velocity[0]*n[0] + \
                self.velocity[1]*n[1] + \
                self.velocity[2]*n[2]

            # Reflect velocity
            self.velocity[0] -= 2 * vn * n[0]
            self.velocity[1] -= 2 * vn * n[1]
            self.velocity[2] -= 2 * vn * n[2]


class PhysicsObjectHandler:
    def __init__(self):
        self.shapes: list[ShapePhysicsObject] = []
    
    def add_shape(self, shape: ShapePhysicsObject):
        self.shapes.append(shape)

    def handle_shapes(self):
        for shape in self.shapes:
            #print(shape.shape)
            shape.update_velocity_from_forces()
            shape.update_pos_from_velocity()
            shape.detect_collision(self.shapes)
            shape.add_shape_to_renderer()
            shape.add_color_to_renderer()
            shape.forces = [0,0,0]





"""
Aiden3DRenderer - A lightweight 3D wireframe renderer

Example usage:
    from aiden3drenderer import Renderer3D, register_shape
    
    renderer = Renderer3D()
    renderer.run()
"""

__version__ = "1.1.1"
__author__ = "Aiden"

from .renderer import Renderer3D, register_shape
from .camera import Camera
from . import shapes
from . import physics

__all__ = [
    "Renderer3D",
    "register_shape",
    "Camera",
    "shapes",
    "physics",
]



"""
Camera controls for 3D renderer
"""
import pygame
import math


class Camera:
    
    def __init__(self, position=None, rotation=None):
        self.position = position if position else [40, 10, 40]
        self.rotation = rotation if rotation else [0, 0, 0]  # pitch, yaw, roll
        
        self.speed = 0.1
        self.base_speed = 0.1
        self.speed_mult = 2
        
        self.mouse_start_pos = None
        self.mouse_start_rotation = None
        self.holding_mouse = False
        
    def handle_mouse_events(self, event):
        if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 3:  # Right mouse button
                self.mouse_start_pos = event.pos
                self.mouse_start_rotation = self.rotation[:]
                self.holding_mouse = True
                
        elif event.type == pygame.MOUSEBUTTONUP:
            if event.button == 3:
                self.holding_mouse = False
    
    def update_mouse_look(self):
        if self.holding_mouse and self.mouse_start_pos:
            current_pos = pygame.mouse.get_pos()
            x_diff = current_pos[0] - self.mouse_start_pos[0]
            y_diff = current_pos[1] - self.mouse_start_pos[1]
            
            self.rotation[0] = self.mouse_start_rotation[0] + y_diff * 0.002  # Pitch
            self.rotation[1] = self.mouse_start_rotation[1] + x_diff * 0.002  # Yaw
    
    def update(self, keys):
        self.update_mouse_look()
        
        forward_x = math.sin(self.rotation[1])
        forward_z = -math.cos(self.rotation[1])
        right_x = math.cos(self.rotation[1])
        right_z = math.sin(self.rotation[1])

        if keys[pygame.K_w]:
            self.position[0] -= forward_x * self.speed
            self.position[2] -= forward_z * self.speed
        if keys[pygame.K_s]:
            self.position[0] += forward_x * self.speed
            self.position[2] += forward_z * self.speed
        if keys[pygame.K_a]:
            self.position[0] += right_x * self.speed
            self.position[2] += right_z * self.speed
        if keys[pygame.K_d]:
            self.position[0] -= right_x * self.speed
            self.position[2] -= right_z * self.speed

        if keys[pygame.K_SPACE]:
            self.position[1] += self.speed
        if keys[pygame.K_LSHIFT]:
            self.position[1] -= self.speed

        if keys[pygame.K_DOWN]:
            self.rotation[0] -= 0.001  # Pitch down
        if keys[pygame.K_UP]:
            self.rotation[0] += 0.001  # Pitch up
        if keys[pygame.K_RIGHT]:
            self.rotation[1] += 0.001  # Yaw right
        if keys[pygame.K_LEFT]:
            self.rotation[1] -= 0.001  # Yaw left
        
        if keys[pygame.K_LCTRL]:
            self.speed = self.base_speed * self.speed_mult
        else:
            self.speed = self.base_speed



"""
Built-in 3D shapes for Aiden3DRenderer
"""
import math
import random
import pygame
from .renderer import register_shape


@register_shape("mountain", pygame.K_1, is_animated=False, color=(150, 0, 150))
def generate_mountain(grid_size=20):
    """Generate a mountain terrain"""
    gridCoords = []
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            center_x = grid_size / 2
            center_z = grid_size / 2
            dist = math.sqrt((x - center_x)**2 + (z - center_z)**2)
            
            max_height = 8
            mountain_radius = grid_size / 3
            if dist < mountain_radius:
                y = max_height * (1 - (dist / mountain_radius)**2)
            else:
                y = 0
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("waves", pygame.K_2, is_animated=True, color=(150, 0, 150))
def generate_waves(grid_size=30, time=0):
    """Generate animated sine wave terrain"""
    gridCoords = []
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            # Multiple sine waves for interesting patterns
            y = math.sin(x * 0.5 + time) * 2
            y += math.cos(z * 0.5 + time) * 2
            y += math.sin((x + z) * 0.3 + time * 0.5) * 1.5
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("ripple", pygame.K_3, is_animated=True, color=(150, 0, 150))
def generate_ripple(grid_size=25, time=0):
    """Generate ripple effect from center"""
    gridCoords = []
    center_x = grid_size / 2
    center_z = grid_size / 2
    
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            dist = math.sqrt((x - center_x)**2 + (z - center_z)**2)
            y = math.sin(dist - time * 3) * 3 * math.exp(-dist / 10)
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("canyon", pygame.K_4, is_animated=False, color=(150, 0, 150))
def generate_canyon(grid_size=30):
    """Generate a canyon/valley terrain"""
    gridCoords = []
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            # Create valley in the middle
            center_x = grid_size / 2
            dist_from_center_x = abs(x - center_x)
            
            # U-shaped valley
            y = -((dist_from_center_x / 5) ** 2) + 5
            
            # Add some noise
            y += math.sin(z * 0.5) * 0.5
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("pyramid", pygame.K_5, is_animated=False, color=(150, 0, 150))
def generate_pyramid(grid_size=15):
    """Generate a stepped pyramid"""
    gridCoords = []
    center = grid_size / 2
    
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            # Distance from center (chebyshev distance for square pyramid)
            dist = max(abs(x - center), abs(z - center))
            
            # Step height based on distance
            y = max(0, (grid_size / 2) - dist)
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("spiral", pygame.K_6, is_animated=True, color=(150, 0, 150))
def generate_spiral(grid_size=25, time=0):
    """Generate a spiral shape"""
    gridCoords = []
    center_x = grid_size / 2
    center_z = grid_size / 2
    
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            dx = x - center_x
            dz = z - center_z
            
            # Convert to polar coordinates
            angle = math.atan2(dz, dx)
            dist = math.sqrt(dx**2 + dz**2)
            
            # Spiral formula
            y = math.sin(angle * 3 + dist * 0.5 + time) * 3
            
            row.append((x, y, z))
        gridCoords.append(row)
    return gridCoords


@register_shape("torus", pygame.K_7, is_animated=False, color=(150, 0, 150))
def generate_torus(resolution=30):
    """Generate a 3D torus (donut shape)"""
    gridCoords = []
    R = 5  # Major radius
    r = 2  # Minor radius
    
    for u_idx in range(resolution):
        row = []
        for v_idx in range(resolution):
            u = (u_idx / resolution) * 2 * math.pi
            v = (v_idx / resolution) * 2 * math.pi
            
            x = (R + r * math.cos(v)) * math.cos(u)
            y = r * math.sin(v)
            z = (R + r * math.cos(v)) * math.sin(u)
            
            row.append((x + 15, y + 5, z + 15))
        gridCoords.append(row)
    return gridCoords


@register_shape("sphere", pygame.K_8, is_animated=False, color=(150, 0, 150))
def generate_sphere(resolution=20, radius=5):
    """Generate a sphere"""
    gridCoords = []
    
    for theta_idx in range(resolution+1):
        row = []
        for phi_idx in range(resolution):
            theta = (theta_idx / resolution) * math.pi
            phi = (phi_idx / resolution) * 2 * math.pi
            
            x = radius * math.sin(theta) * math.cos(phi)
            y = radius * math.cos(theta)
            z = radius * math.sin(theta) * math.sin(phi)
            
            row.append((x, y, z))
        row.append(row[0])
        gridCoords.append(row)
    return gridCoords


@register_shape("mobius", pygame.K_9, is_animated=False, color=(150, 0, 150))
def generate_mobius_strip(resolution=30):
    """Generate a Möbius strip"""
    gridCoords = []
    
    for u_idx in range(resolution):
        row = []
        for v_idx in range(resolution):
            u = (u_idx / resolution) * 2 * math.pi
            v = (v_idx / resolution - 0.5) * 2  # -1 to 1
            
            # Möbius strip parametric equations
            x = (2 + v * math.cos(u / 2)) * math.cos(u)
            y = v * math.sin(u / 2)
            z = (2 + v * math.cos(u / 2)) * math.sin(u)
            
            row.append((x + 15, y + 5, z + 15))
        gridCoords.append(row)
    return gridCoords


@register_shape("megacity", pygame.K_0, is_animated=False, color=(150, 0, 150))
def generate_megacity(grid_size=80):
    """Generate a massive futuristic city with procedural buildings"""
    gridCoords = []
    random.seed(42)  # Consistent generation
    
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            # Create city blocks
            block_x = x // 8
            block_z = z // 8
            in_block_x = x % 8
            in_block_z = z % 8
            
            # Roads (empty spaces between blocks)
            is_road = (in_block_x == 0 or in_block_x == 7 or in_block_z == 0 or in_block_z == 7)
            
            if is_road:
                y = 0
            else:
                # Generate building heights with variation
                building_seed = block_x * 1000 + block_z
                random.seed(building_seed)
                
                # Distance from center affects building height
                center = grid_size / 2
                dist_from_center = math.sqrt((x - center)**2 + (z - center)**2)
                center_factor = max(0, 1 - dist_from_center / (grid_size / 2))
                
                # Random building height (taller in center)
                base_height = random.random() * 25 * center_factor + 3
                
                # Add some variation within the building
                variation = math.sin(in_block_x * 2) * math.cos(in_block_z * 2) * 0.5
                
                # Some buildings have antennas
                if random.random() > 0.7 and in_block_x == 3 and in_block_z == 3:
                    y = base_height + 5
                else:
                    y = base_height + variation
            
            row.append((x, y, z))
        gridCoords.append(row)
    
    random.seed()  # Reset random seed
    return gridCoords


@register_shape("alien", pygame.K_q, is_animated=True, color=(150, 0, 150))
def generate_alien_landscape(grid_size=60, time=0):
    """Generate an otherworldly alien landscape with multiple features"""
    gridCoords = []
    
    for x in range(grid_size):
        row = []
        for z in range(grid_size):
            # Multiple overlapping features
            center_x = grid_size / 2
            center_z = grid_size / 2
            
            # Feature 1: Large crater
            crater_dist = math.sqrt((x - center_x)**2 + (z - center_z)**2)
            crater = 0
            if crater_dist < 20:
                crater = -(crater_dist - 20)**2 / 40 + 10
            
            # Feature 2: Crystalline spikes
            spike_pattern = abs(math.sin(x * 0.8) * math.cos(z * 0.8))
            spikes = spike_pattern ** 3 * 15
            
            # Feature 3: Rolling hills
            hills = math.sin(x * 0.3) * 3 + math.cos(z * 0.4) * 2.5
            
            # Feature 4: Alien vegetation (thin tall structures)
            veg_noise = (math.sin(x * 2.3 + z * 1.7) + 1) / 2
            if veg_noise > 0.85:
                vegetation = (veg_noise - 0.85) * 40
            else:
                vegetation = 0
            
            # Feature 5: Pulsating energy field
            pulse_dist = math.sqrt((x - center_x + 15)**2 + (z - center_z + 15)**2)
            pulse = math.sin(pulse_dist * 0.5 - time * 2) * 2 * math.exp(-pulse_dist / 20)
            
            # Combine all features
            y = crater + spikes * 0.3 + hills + vegetation + pulse
            
            row.append((x, y, z))
        gridCoords.append(row)
    
    return gridCoords


@register_shape("helix", pygame.K_e, is_animated=True, color=(150, 0, 150))
def generate_double_helix(length=60, time=0):
    """Generate a DNA-like double helix structure"""
    gridCoords = []
    
    for i in range(length):
        row = []
        for j in range(20):
            angle = (j / 10) * math.pi  # 0 to 2π
            z_pos = i
            
            # Two intertwined helices
            if j < 10:
                # First strand
                radius = 5
                x = radius * math.cos(angle + z_pos * 0.3 + time)
                y = radius * math.sin(angle + z_pos * 0.3 + time)
            else:
                # Second strand (opposite phase)
                radius = 5
                x = radius * math.cos(angle + z_pos * 0.3 + time + math.pi)
                y = radius * math.sin(angle + z_pos * 0.3 + time + math.pi)
            
            row.append((x + 30, y + 10, z_pos))
        gridCoords.append(row)
    
    return gridCoords


@register_shape("mandelbulb", pygame.K_r, is_animated=False, color=(150, 0, 150))
def generate_mandelbulb_slice(resolution=50, z_slice=0, max_iterations=10):
    """Generate a 2D slice of a 3D Mandelbulb fractal"""
    gridCoords = []
    
    for x_idx in range(resolution):
        row = []
        for y_idx in range(resolution):
            # Map to Mandelbulb coordinate space
            x = (x_idx / resolution - 0.5) * 3
            y = (y_idx / resolution - 0.5) * 3
            z = z_slice
            
            # Calculate Mandelbulb iterations
            x0, y0, z0 = x, y, z
            iteration = 0
            
            for _ in range(max_iterations):
                r = math.sqrt(x**2 + y**2 + z**2)
                if r > 2:
                    break
                
                theta = math.atan2(math.sqrt(x**2 + y**2), z)
                phi = math.atan2(y, x)
                
                # Power 8 Mandelbulb formula
                r8 = r ** 8
                x = r8 * math.sin(8 * theta) * math.cos(8 * phi) + x0
                y = r8 * math.sin(8 * theta) * math.sin(8 * phi) + y0
                z = r8 * math.cos(8 * theta) + z0
                
                iteration += 1
            
            # Use iteration count as height
            height = iteration * 2
            
            row.append((x_idx, height, y_idx))
        gridCoords.append(row)
    
    return gridCoords


@register_shape("klein", pygame.K_t, is_animated=False, color=(150, 0, 150))
def generate_klein_bottle(resolution=40):
    """Generate a Klein bottle - a non-orientable surface"""
    gridCoords = []
    
    for u_idx in range(resolution):
        row = []
        for v_idx in range(resolution):
            u = (u_idx / resolution) * 2 * math.pi
            v = (v_idx / resolution) * 2 * math.pi
            
            # Klein bottle parametric equations
            r = 4 * (1 - math.cos(u) / 2)
            
            if u < math.pi:
                x = 6 * math.cos(u) * (1 + math.sin(u)) + r * math.cos(u) * math.cos(v)
                y = 16 * math.sin(u) + r * math.sin(u) * math.cos(v)
            else:
                x = 6 * math.cos(u) * (1 + math.sin(u)) + r * math.cos(v + math.pi)
                y = 16 * math.sin(u)
            
            z = r * math.sin(v)
            
            row.append((x + 30, y + 20, z + 30))
        gridCoords.append(row)
    
    return gridCoords


@register_shape("trefoil", pygame.K_y, is_animated=False, color=(150, 0, 150))
def generate_trefoil_knot(resolution=50):
    """Generate a trefoil knot - a classic mathematical knot"""
    gridCoords = []
    
    for t_idx in range(resolution):
        row = []
        for r_idx in range(15):
            t = (t_idx / resolution) * 2 * math.pi
            
            # Trefoil knot parametric equations
            x = math.sin(t) + 2 * math.sin(2 * t)
            y = math.cos(t) - 2 * math.cos(2 * t)
            z = -math.sin(3 * t)
            
            # Add thickness
            theta = (r_idx / 15) * 2 * math.pi
            radius = 0.5
            
            # Normal vector calculation (simplified)
            nx = math.cos(t)
            ny = math.sin(t)
            
            x_thick = x + radius * math.cos(theta) * nx
            y_thick = y + radius * math.cos(theta) * ny
            z_thick = z + radius * math.sin(theta)
            
            row.append((x_thick * 5 + 25, y_thick * 5 + 15, z_thick * 5 + 25))
        gridCoords.append(row)
    
    return gridCoords


@register_shape("plane", None, False)
def generate_plane(size=20, rot_x=0, rot_y=0, rot_z=0):
    """Generate a flat plane centered at (0,0,0), optionally rotated around x, y, z axes (in radians)"""
    grid_coords = []
    cx = size / 2
    cz = size / 2
    for z in range(size):
        yL = []
        for x in range(size):
            px = x - cx
            py = 0
            pz = z - cz

            # Rotation around X axis
            y1 = py * math.cos(rot_x) - pz * math.sin(rot_x)
            z1 = py * math.sin(rot_x) + pz * math.cos(rot_x)
            x1 = px

            # Rotation around Y axis
            x2 = x1 * math.cos(rot_y) + z1 * math.sin(rot_y)
            z2 = -x1 * math.sin(rot_y) + z1 * math.cos(rot_y)
            y2 = y1

            # Rotation around Z axis
            x3 = x2 * math.cos(rot_z) - y2 * math.sin(rot_z)
            y3 = x2 * math.sin(rot_z) + y2 * math.cos(rot_z)
            z3 = z2

            yL.append((x3, y3, z3))
        grid_coords.append(yL)

    return grid_coords


"""
Main 3D renderer class
"""
import math
import pygame
from pygame import QUIT
import sys

from .camera import Camera

CUSTOM_SHAPES = {}

def register_shape(name: str, key=None, is_animated: bool = False, color: tuple[int] = None):
    def decorator(func):
        CUSTOM_SHAPES[name] = {
            'function': func,
            'is_animated': is_animated,
            'key': key,
            'color': color
        }
        return func
    return decorator


class Renderer3D:
    
    def __init__(self, width=1000, height=1000, title="Aiden 3D Renderer"):
        pygame.init()
        self.width = width
        self.height = height
        self.half_w = width // 2
        self.half_h = height // 2
        self.screen = pygame.display.set_mode((width, height))
        pygame.display.set_caption(title)
        
        self.camera = Camera()
        self.clock = pygame.time.Clock()
        
        self.current_shape = "mountain"
        self.last_shape = None
        self.animation_time = 0
        self.grid_coords = None
        self.needs_regen = False
        self.shapes = [self.current_shape]
        self.projected = None

        self.is_starting = True

        self.is_mesh = True
        self.triangle_base_color_1= (150, 0, 150)
        self.triangle_base_color_2 = (50, 0, 50)
        self.triangle_color_list_1= []
        self.triangle_color_list_2 = []

        self.grid_coords_list = []
        self.projections_list = []

    def set_starting_shape(self, shape: str):
        self.current_shape = shape
        self.shapes = [self.current_shape]

    def project_3d_to_2d(self, matrix, fov, camera_pos, camera_facing):
        projected = []
        #print(len(self.grid_coords_list))
        if matrix is None:
            return None
        for xIdx in range(len(matrix)):
            xList = matrix[xIdx]
            row = []
            for yIdx in range(len(xList)):
                point = xList[yIdx]

                if point is None:
                    row.append(None)
                    continue

                x = point[0]
                y = point[1]
                z = point[2]

                x -= camera_pos[0]
                y -= camera_pos[1]
                z -= camera_pos[2]

                x1 = x * math.cos(camera_facing[1]) + z * math.sin(camera_facing[1])
                y1 = y
                z1 = -x * math.sin(camera_facing[1]) + z * math.cos(camera_facing[1])

                x2 = x1
                y2 = y1 * math.cos(camera_facing[0]) - z1 * math.sin(camera_facing[0])
                z2 = y1 * math.sin(camera_facing[0]) + z1 * math.cos(camera_facing[0])

                x3 = x2 * math.cos(camera_facing[2]) - y2 * math.sin(camera_facing[2])
                y3 = x2 * math.sin(camera_facing[2]) + y2 * math.cos(camera_facing[2])
                z3 = z2

                if z3 <= 0.1:
                    row.append(None)
                    continue

                f = 1 / math.tan(fov / 2)
                dd_x = (x3 * f) / -z3
                dd_y = (y3 * f) / -z3

                px = dd_x * self.half_w + self.half_w
                py = dd_y * self.half_h + self.half_h

                margin = 1000
                if px < -margin or px > self.width + margin or py < -margin or py > self.height + margin:
                    row.append(None)
                    continue
                
                if self.is_mesh:
                    row.append((px, py))
                else:
                    row.append((px, py, z3))
            projected.append(row)

        return projected
    
    def render_wireframe(self, matrix):
        if self.is_mesh:
            if matrix is None:
                return
            for xIdx in range(len(matrix)):
                xList = matrix[xIdx]
                for yIdx in range(len(xList)):
                    point = xList[yIdx]
                    if point is not None:
                        points = []

                        try:
                            if xIdx < len(matrix) - 1 and matrix[xIdx+1][yIdx] is not None:
                                points.append(matrix[xIdx+1][yIdx])
                            if yIdx < len(xList) - 1 and matrix[xIdx][yIdx+1] is not None:
                                points.append(matrix[xIdx][yIdx+1])
                        except IndexError:
                            continue

                        for p in points:
                            pygame.draw.line(self.screen, (0, 0, 0), point, p, 2)
        else:
            all_tris = []
            for matI in range(len(matrix)):
                mat = matrix[matI]
                if mat is None:
                    continue
                tris = []

                #print(len(self.triangle_color_list_1))

                col1 = self.triangle_color_list_1[matI] if self.triangle_color_list_1[matI] is not None else self.triangle_base_color_1
                col2 = self.triangle_color_list_2[matI] if self.triangle_color_list_2[matI] is not None else self.triangle_base_color_2
                
                for xIdx in range(len(mat)):
                    xList = mat[xIdx]
                    for yIdx in range(len(xList)):
                        point = xList[yIdx]
                        if point is not None:

                            if xIdx < len(mat) - 1 and yIdx < len(xList) - 1:
                                p1 = mat[xIdx][yIdx + 1]
                                p2 = mat[xIdx + 1][yIdx]
                                if p1 is not None and p2 is not None:
                                    #pygame.draw.polygon(screen, (150, 0, 150), [point, p1, p2], 0)
                                    d1 = (point[2] + p1[2] + p2[2]) / 3 if len(point) > 2 else 0
                                    tris.append((d1, (point, p1, p2), col1))

                            if xIdx > 0 and yIdx > 0:
                                p1 = mat[xIdx][yIdx - 1]
                                p2 = mat[xIdx - 1][yIdx]
                                if p1 is not None and p2 is not None:
                                    #pygame.draw.polygon(screen, (50, 0, 50), [point, p1, p2], 0)
                                    d1 = (point[2] + p1[2] + p2[2]) / 3 if len(point) > 2 else 0
                                    tris.append((d1, (point, p1, p2), col2))
                all_tris.extend(tris)
            
            all_tris.sort(key=lambda t: t[0], reverse=True)

            for _, tri, col in all_tris:
                pygame.draw.polygon(
                    self.screen,
                    col,
                    [(tri[0][0], tri[0][1]), (tri[1][0], tri[1][1]), (tri[2][0], tri[2][1])],
                    0,
                )
    
    def generate_shape(self, shape_name, time=0):
        if shape_name in CUSTOM_SHAPES:
            shape_info = CUSTOM_SHAPES[shape_name]
            func = shape_info['function']
            color = shape_info['color']

            if color is not None:
                self.triangle_color_list_2.append((color[0]*0.75, color[1]*0.75, color[2]*0.75))
            else:
                self.triangle_color_list_2.append(None)

            self.triangle_color_list_1.append(color)
            
            
            if shape_info['is_animated']:
                return func(time=time), True
            else:
                return func(), False
        
        return None, False
    
    def generate_shape_from_key_press(self, pressedKeys, time=0):
        shapesList = []
        for shape, value in CUSTOM_SHAPES.items():
            key = value['key']
            if key is not None:
                if pressedKeys[key]:
                    shapesList.append(shape)
        
        if len(shapesList) >= 1:
            self.shapes = shapesList
        
    def min_z(self, lvl1):
        values = [
            t[2]
            for lvl2 in lvl1
            for t in lvl2
            if t is not None
        ]
        return max(values) if values else float("-inf")

    def run(self):
        running = True
        while running:
            self.screen.fill((255, 255, 255))
            self.clock.tick(60)
            self.animation_time += 0.01
            
            # Handle events
            for event in pygame.event.get():
                if event.type == QUIT:
                    running = False
                    
                self.camera.handle_mouse_events(event)
            
            keys = pygame.key.get_pressed()
            if keys[pygame.K_ESCAPE]:
                running = False
            
            self.camera.update(keys)
            
            self.generate_shape_from_key_press(keys, self.animation_time)

            for i in range(len(self.shapes)):
                shape_name = self.shapes[i]
                #print(shape_name)
                self.grid_coords_list.append(self.generate_shape(shape_name, self.animation_time))

            for i in range(len(self.grid_coords_list)):
                self.grid_coords = self.grid_coords_list[i]
                if self.grid_coords:
                    projected = self.project_3d_to_2d(
                        self.grid_coords[0],
                        math.radians(100),
                        tuple(self.camera.position),
                        tuple(self.camera.rotation)
                    )
                    self.projections_list.append(projected)

            if not self.is_mesh:
                self.render_wireframe(self.projections_list)
            else:
                for proj in self.projections_list:
                    self.render_wireframe(proj)

            self.grid_coords_list = []
            self.triangle_color_list_1 = []
            self.triangle_color_list_2 = []
            self.projections_list = []


            pygame.display.update()
        
        pygame.quit()

    def loopable_run(self):
        self.screen.fill((255, 255, 255))
        self.clock.tick(60)
        self.animation_time += 0.01
        
        # Handle events
        for event in pygame.event.get():
            if event.type == QUIT:
                pygame.quit()
                
            self.camera.handle_mouse_events(event)
        
        keys = pygame.key.get_pressed()
        if keys[pygame.K_ESCAPE]:
            pygame.quit()
        
        self.camera.update(keys)
        
        self.generate_shape_from_key_press(keys, self.animation_time)

        for i in range(len(self.shapes)):
            shape_name = self.shapes[i]
            #print(shape_name)
            self.grid_coords_list.append(self.generate_shape(shape_name, self.animation_time))

        for i in range(len(self.grid_coords_list)):
            self.grid_coords = self.grid_coords_list[i]
            if self.grid_coords:
                projected = self.project_3d_to_2d(
                    self.grid_coords[0],
                    math.radians(100),
                    tuple(self.camera.position),
                    tuple(self.camera.rotation)
                )
                self.projections_list.append(projected)
        
        if not self.is_mesh:
            self.render_wireframe(self.projections_list)
        else:
            for proj in self.projections_list:
                self.render_wireframe(proj)

        self.grid_coords_list = []
        self.triangle_color_list_1 = []
        self.triangle_color_list_2 = []
        self.projections_list = []
        
        pygame.display.update()
        


def main():
    renderer = Renderer3D()
    renderer.run()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="aiden3drenderer",
    version="1.1.1",
    author="Aiden",
    author_email="headstone.yt@gmail.com",
    description="A lightweight 3D wireframe renderer built from scratch using Pygame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AidenKielby/3D-mesh-Renderer",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    python_requires=">=3.9",
    install_requires=[
        "pygame>=2.6.0",
    ],
    extras_require={
        "dev": ["pytest>=7.0", "black", "flake8"],
    },
    entry_points={
        "console_scripts": [
            "aiden3d-demo=aiden3drenderer.renderer:main",
        ],
    },
    include_package_data=True,

)